"""The base classes and abstract classes for working with :class:`Transformation`."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Optional, Sequence, TypeVar, Union

from spekk import Spec
from spekk.transformations.common import get_fn_name

TBuildableSelf = TypeVar("TBuildableSelf", bound="Buildable")


class Buildable(ABC):
    """An abstract class representing something that can be "built" with a
    :class:`Spec` to get information about how the spec changes when transforming
    functions.

    Args:
        input_spec (Spec): The spec that is passed into the function.
        passed_spec (Spec): The spec that is passed down into the wrapped function
            after the transformation has been applied.
        returned_spec (Spec): The spec that is returned up from the wrapped function.
        output_spec (Spec): The final spec that is returned after calling the
            transformed function.
    """

    input_spec: Optional[
        Spec
    ] = None  #: The spec for the input kwargs for the function.
    passed_spec: Optional[
        Spec
    ] = None  #: The spec for the input kwargs that are passed down into the wrapped function after the transformation has been applied.
    returned_spec: Optional[
        Spec
    ] = None  #: The spec of the returned value from the wrapped function.
    output_spec: Optional[
        Spec
    ] = None  #: The spec of the final returned value from the transformed function.

    @abstractmethod
    def build(self: TBuildableSelf, input_spec: Spec) -> TBuildableSelf:
        """Return a copy of this object with information about what happens to the spec
        when the function is called."""

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            + f"input_spec={self.input_spec}, "
            + f"passed_spec={self.passed_spec}, "
            + f"returned_spec={self.returned_spec}, "
            + f"output_spec={self.output_spec})"
        )


def _transformed_function_repr_fn(
    tf: "TransformedFunction",
    step_repr_fn: Optional[Callable[["TransformedFunction", str], str]] = None,
):
    "Format a TransformedFunction in a nice way."
    s = f"TransformedFunction(\n  <compose(\n"
    for step in tf.traverse(depth_first=True):
        step_repr = (
            repr(step.transformation)
            if isinstance(step, TransformedFunction)
            else get_fn_name(step)
        )
        if step_repr_fn is not None:
            step_repr = step_repr_fn(step, step_repr)
        else:
            step_repr = f"    {step_repr},\n"
        s += step_repr
    s += "  )>\n)"
    return s


@dataclass
class TransformedFunctionError(Exception):
    """An error that occurred while calling a :class:`TransformedFunction`.

    It will print out the nested steps vertically and highlight the step that raised an
    exception along with the original exception."""

    original_exception: Exception
    transformed_function: "TransformedFunction"
    error_step: Union[callable, "TransformedFunction"]

    def __post_init__(self):
        if isinstance(self.transformed_function, TransformedFunction):
            s = _transformed_function_repr_fn(
                self.transformed_function,
                lambda step, default_repr: f"⚠   {default_repr},\n"
                + f"⚠     ↳ This step raised {repr(self.original_exception)}\n"
                if step is self.error_step
                else f"    {default_repr},\n",
            )
            s = f"{str(self.original_exception)}\n" + s
        else:
            s = str(self.original_exception)
        super().__init__(s)


@dataclass
class _WrappedWithErrorHandling:
    """A wrapper around a function that catches errors and re-raises them as
    :class:`TransformedFunctionError`s.

    It is important to re-raise exceptions as :class:`TransformedFunctionError`s
    because then we can give users more information about where the error occurred in
    a sequence of transformations.

    See also:
        :class:`TransformedFunctionError`"""

    f: callable

    def __call__(self, *args, **kwargs):
        try:
            return self.f(*args, **kwargs)
        except Exception as e:
            raise TransformedFunctionError(e, self.f, self.f) from e


class _NoSpecGiven(Spec):
    """A dummy spec that raises an error when it is used.

    It is used to get better error messages when calling a :class:`TransformedFunction`
    that hasnot been built with a :class:`Spec`."""

    def __getattribute__(self, _: str) -> Any:
        raise ValueError(
            f"The Transformation tried to use the spec, but no spec was given. Did you \
forget to call build()?"
        )


@dataclass
class TransformedFunction(Buildable):
    wrapped_fn: Union[
        callable, "TransformedFunction"
    ]  #: The original function that was wrapped by the transformation.
    transformation: "Transformation"  #: The :class:`Transformation` that was applied to the wrapped function.

    def __call__(self, *args, **kwargs):
        try:
            # Handle the case where the function has not been built yet. If any
            # transformation requires/uses a spec, and this object has not been built
            # with a spec, it will raise an error.
            input_spec = self.input_spec
            returned_spec = self.returned_spec
            if input_spec is None:
                input_spec = _NoSpecGiven()
            if returned_spec is None:
                returned_spec = _NoSpecGiven()

            # Handle special case where the wrapped function is not a
            # TransformedFunction (e.g. it's the kernel function)
            wrapped_fn = self.wrapped_fn
            if not isinstance(wrapped_fn, TransformedFunction):
                wrapped_fn = _WrappedWithErrorHandling(wrapped_fn)

            # Perform the actual transformation
            transformed_wrapped_function = self.transformation.transform_function(
                wrapped_fn, input_spec, returned_spec
            )

            # Return the result of calling the transformed function
            return transformed_wrapped_function(*args, **kwargs)

        # Handle errors that can occur while running the transformed function
        except TransformedFunctionError as e:
            # A nested step raised an error. Let's add this step to the stack trace
            raise TransformedFunctionError(
                e.original_exception, self, e.error_step
            ) from e.original_exception
        except Exception as e:
            # An exception was raised in this step.
            raise TransformedFunctionError(e, self, self) from e

    def build(self, input_spec: Spec) -> "TransformedFunction":
        try:
            # The spec that will be passed into the wrapped function:
            passed_spec = self.transformation.transform_input_spec(input_spec)

            # Recursively build the wrapped function
            wrapped_fn = self.wrapped_fn
            if isinstance(wrapped_fn, Buildable):
                wrapped_fn = wrapped_fn.build(passed_spec)
                # The spec after the wrapped function has been called:
                returned_spec = wrapped_fn.output_spec
            else:
                # By default, we assume that the wrapped function returns a scalar.
                returned_spec = Spec(())

            # Lastly, transform the output spec according to the Transformation.
            output_spec = self.transformation.transform_output_spec(returned_spec)

            # Return a new copy with the updated specs
            copy = TransformedFunction(wrapped_fn, self.transformation)
            copy.input_spec = input_spec
            copy.passed_spec = passed_spec
            copy.returned_spec = returned_spec
            copy.output_spec = output_spec
            return copy

        # Handle errors that can occur while building
        except TransformedFunctionError as e:
            # A nested step raised an error. Let's reraise the exception.
            raise TransformedFunctionError(
                e.original_exception, self, e.error_step
            ) from e.original_exception
        except Exception as e:
            # An exception was raised in this step.
            raise TransformedFunctionError(e, self, self) from e

    def traverse(self, *, depth_first: bool = False):
        """Recursively yield the potentially nested :class:`TransformedFunction`.

        >>> from spekk.transformations import ForAll, compose, TransformedFunction
        >>> tf = compose(abs, ForAll("x"), ForAll("y"))
        >>> for t in tf.traverse():
        ...     if isinstance(t, TransformedFunction):
        ...         t = t.transformation  # Print transformation, not TransformedFunction
        ...     print(repr(t))
        ForAll("y")
        ForAll("x")
        <built-in function abs>
        """
        if not depth_first:
            yield self
        if isinstance(self.wrapped_fn, TransformedFunction):
            yield from self.wrapped_fn.traverse(depth_first=depth_first)
        else:
            yield self.wrapped_fn
        if depth_first:
            yield self

    def __repr__(self):
        return _transformed_function_repr_fn(self)

    def __hash__(self):
        return object.__hash__(self)


class Transformation(ABC):
    """A :class:`Transformation` takes a function and transforms/modifies it. It also
    keeps track of changes to the spec when transforming the function. See the
    schematic in :mod:`spekk.transformations`'s docstring for a visualization of how specs are processed.

    A :class:`Transformation` acts as a higher-order function that takes in a function
    and returns an instance of :class:`TransformedFunction`. It gets information about
    the dimensions of the data (through instances of :class:`Spec`) that the function
    operates on and it also defines what happens to the spec when the function is
    transformed.

    Attributes:
        transform_function: The main transformation function that returns a
            :class:`TransformedFunction`.
        transform_input_spec: A function that returns the spec that is passed down into
            the wrapped function after the transformation has been applied.
        transform_output_spec: A function that returns the final returned spec of the
            transformed function.
    """

    def __call__(self, wrapped_fn: callable) -> TransformedFunction:
        "Transform the wrapped function."
        if isinstance(wrapped_fn, PartialTransformation):
            return PartialTransformation([*wrapped_fn.partial_transformations, self])
        elif isinstance(wrapped_fn, Transformation):
            # Handle partial application of transformations.
            return PartialTransformation([wrapped_fn, self])
        return TransformedFunction(wrapped_fn, self)

    @abstractmethod
    def transform_function(
        self, wrapped_fn: callable, input_spec: Spec, output_spec: Spec
    ) -> TransformedFunction:
        """Transform the wrapped function given the spec of the input arguments and the
        spec of the returned value of the wrapped function.
        """

    @abstractmethod
    def transform_input_spec(self, spec: Spec) -> Spec:
        """Return a new spec that represent the input arguments that are passed down to
        the wrapped function after the transformation has been applied.

        For example, if the transformation vectorizes the wrapped function over a
        dimension, the wrapped function would only see single items of the dimension at
        a time. Therefore, the input spec will have one less dimension when passed down
        to the wrapped function.
        """

    @abstractmethod
    def transform_output_spec(self, spec: Spec) -> Spec:
        """Return a new spec that represent the returned value of the final transformed
        function.

        For example, if the transformation sums over a dimension of the result of
        calling the wrapped function, the output spec will have one less dimension.
        """


@dataclass
class PartialTransformation(Transformation):
    """A partially applied transformation.

    :class:`PartialTransformation` lets us compose transformations such that they can
    be re-used as building blocks. For example, in the below code, ``tf_partial`` and
    ``tf_full`` are equivalent:

    >>> import numpy as np
    >>> from spekk.transformations import ForAll, compose

    Let's create an example kernel function along with some data and a spec:

    >>> kernel = lambda x: x**2
    >>> data = {"x": np.ones((2, 3)) * 2}
    >>> spec = Spec({"x": ["a", "b"]})

    Create a partial transformation by composing two transformations:

    >>> forall_xy = compose(ForAll("a"), ForAll("b"))

    Create and build two equivalent transformed functions:

    >>> tf_partial = compose(kernel, forall_xy).build(spec)
    >>> tf_full = compose(kernel, ForAll("a"), ForAll("b")).build(spec)

    The two transformed functions are equivalent:

    >>> np.array_equal(tf_partial(**data), tf_full(**data))
    True
    """

    partial_transformations: Sequence[Transformation]

    def __call__(self, wrapped_fn: callable) -> TransformedFunction:
        for t in self.partial_transformations:
            if isinstance(t, PartialTransformation):
                # Makes sure that nested partial transformations are flattened
                wrapped_fn = t(wrapped_fn)
            else:
                wrapped_fn = TransformedFunction(wrapped_fn, t)
        return wrapped_fn

    def transform_function(
        self, wrapped_fn: callable, input_spec: Spec, output_spec: Spec
    ) -> callable:
        "Transform the wrapped function by applying each partial transformation in turn."
        for t in self.partial_transformations:
            wrapped_fn = t.transform_function(wrapped_fn, input_spec, output_spec)
            input_spec = t.transform_input_spec(input_spec)
            output_spec = t.transform_output_spec(output_spec)
        return wrapped_fn

    def transform_input_spec(self, spec: Spec) -> Spec:
        "Transform the input spec using each partial transformation in turn."
        for t in self.partial_transformations:
            spec = t.transform_input_spec(spec)
        return spec

    def transform_output_spec(self, spec: Spec) -> Spec:
        "Transform the output spec using each partial transformation in turn."
        for t in self.partial_transformations:
            spec = t.transform_output_spec(spec)
        return spec

    def __repr__(self):
        return f"PartialTransformation({self.partial_transformations})"


@dataclass
class Specced(Buildable):
    """A wrapper around ``f`` that has information about what happens to the spec when
    the function is called."""

    f: callable  #: The function to wrap.
    transform_spec: Callable[
        [Spec], Spec
    ]  #: A function that transforms the spec, returning the spec of the return-value of ``f``.

    def build(self, input_spec: Spec) -> "Specced":
        output_spec = self.transform_spec(input_spec)
        if not isinstance(output_spec, Spec):
            output_spec = Spec(output_spec)

        new_obj = Specced(self.f, self.transform_spec)
        new_obj.input_spec = input_spec
        new_obj.output_spec = output_spec
        return new_obj

    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)

    def __hash__(self):
        return hash((self.f, self.transform_spec))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import os
import sys

# Add spekk to path
sys.path.insert(0, os.path.abspath(".."))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "spekk"
copyright = "2023, Magnus Dalen Kvalevåg"
author = "Magnus Dalen Kvalevåg"
release = "1.0.3"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",  # Generates documentation from docstrings
    "sphinx.ext.napoleon",  # Adds support for NumPy and Google style docstrings
    "sphinx.ext.autosummary",  # Automatically generates documentation for modules
    "sphinx.ext.viewcode",
    "sphinx_copybutton",  # Adds copy-button to code-blocks
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]



"""An :class:`Axis` references an axis of an array by its corresponding dimension name 
in a spec."""

from dataclasses import dataclass
from functools import reduce
from typing import Sequence, Tuple

import spekk.trees as trees
from spekk.spec import Spec
from spekk.trees import Tree, has_treedef, leaves


@dataclass
class Axis:
    """A placeholder for an array axis, given by the name of that axis (dimension).

    In the context of a transformation, an :class:`Axis` is a way to get the concrete 
    axis-index of an array, and also to specify what happens to that dimension in the
    transformation.

    By default, the dimension is removed, which makes sense for common operations like
    :func:`numpy.sum` or :func:`numpy.mean`. If you want to keep the dimension, set 
    ``keep=True``. If you want to replace the dimension with something else, set 
    ``becomes=("something", "else")``.
    """

    dimension: str  #: The referenced dimension
    keep: bool = False  #: Whether to keep the dimension in the spec after referencing it (default ``False``, i.e. the dimension is removed from the spec).
    becomes: Tuple[str] = ()  #: If set, the dimension is replaced in the spec with the given dimensions.

    def new_dimensions(self, dimensions: Sequence[str]) -> Tuple[str]:
        """Given a sequence of dimensions return the new dimensions after this 
        :class:`Axis` has been parsed.

        By default, the dimension is removed. If ``keep=True``, the dimension is kept. 
        If ``becomes`` is set, the dimension is replaced with the dimensions defined by 
        the ``becomes`` field.

        Examples:

        >>> old_dimensions = ("a", "b", "c")
        >>> Axis("b").new_dimensions(old_dimensions)
        ('a', 'c')
        >>> Axis("b", keep=True).new_dimensions(old_dimensions)
        ('a', 'b', 'c')
        >>> Axis("b", becomes=("x", "y")).new_dimensions(old_dimensions)
        ('a', 'x', 'y', 'c')
        """
        if self.becomes:
            return reduce(
                lambda a, b: (
                    a + tuple(self.becomes) if b == self.dimension else a + (b,)
                ),
                dimensions,
                (),
            )
        elif self.keep:
            return tuple(dimensions)
        else:
            return tuple(d for d in dimensions if d != self.dimension)

    def __repr__(self) -> str:
        repr_str = f'Axis("{self.dimension}"'
        if self.keep:
            repr_str += ", keep=True"
        if self.becomes:
            repr_str += f", becomes={self.becomes}"
        repr_str += ")"
        return repr_str


class AxisConcretizationError(ValueError):
    def __init__(self, axis: Axis):
        super().__init__(f'Could not find dimension "{axis.dimension}" in the spec.')


def concretize_axes(spec: Spec, args: Tree, kwargs: Tree) -> Tuple[list, dict]:
    """Convert any instance of :class:`Axis` in ``args`` and ``kwargs`` to the concrete 
    axis index, as defined by the spec.

    >>> spec = Spec(["a", "b"])
    >>> args = (Axis("a"), Axis("b"))
    >>> kwargs = {"baz": Axis("b")}
    >>> concretize_axes(spec, args, kwargs)
    ((0, 1), {'baz': 1})
    """
    state = (args, kwargs)
    for leaf in leaves(state, lambda x: isinstance(x, Axis) or not has_treedef(x)):
        if isinstance(leaf.value, Axis):
            index = spec.index_for(leaf.value.dimension)
            if index is None:
                raise AxisConcretizationError(leaf.value)
            state = trees.set(state, index, leaf.path)
    return state


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from hypothesis import strategies as st

from spekk import Spec

alphabet = "abcdefghijklmnopqrstuvwxyz"
keys = st.text(alphabet, min_size=1, max_size=3)
dimensions = st.lists(keys, min_size=0, max_size=5, unique=True)


def _specs_tree():
    return st.recursive(
        dimensions,
        lambda children: st.one_of(
            st.lists(children, max_size=3),
            st.dictionaries(keys, children, max_size=3),
        ),
        max_leaves=10,
    )


def specs():
    return _specs_tree().map(Spec)


def kwargs_specs():
    return st.dictionaries(keys, _specs_tree(), max_size=3).map(Spec)



"Utility functions for ``spekk``."

from spekk.util.flatten import flatten
from spekk.util.shape import shape
from spekk.util.slicing import slice_data, slice_spec
from spekk.util.validation import ValidationError, validate

__all__ = [
    "flatten",
    "shape",
    "slice_data",
    "slice_spec",
    "ValidationError",
    "validate",
]



"""Module containing the :class:`Spec` class — the most important component of the 
``spekk`` library."""

from functools import reduce
from typing import Dict, Optional, Sequence, Set, Union

import spekk.trees.core as trees
from spekk.trees import Tree, TreeLens, leaves, register_dispatch_fn, traverse, treedef
from spekk.trees.registry import Tree


def _is_spec_leaf(tree: Optional[Tree]):
    """A Spec-tree is a leaf if it is None or a sequence of strings.

    >>> _is_spec_leaf(None)
    True
    >>> _is_spec_leaf(["a", "b"])
    True

    Anything else is not a leaf, in the following case a list of list of strings:
    >>> _is_spec_leaf([["a"], ["c"]])
    False
    """
    if tree is None:
        return True
    if isinstance(tree, Sequence) and all(isinstance(x, str) for x in tree):
        return True
    if isinstance(tree, Spec):
        return _is_spec_leaf(tree.tree)
    return False


class Spec(TreeLens):
    """In a nested tree of arrays, a Spec describes the dimensions of the arrays. Spec
    is a subclass of :class:`TreeLens` which takes the ``tree`` as an argument when
    constructing an object.

    The tree of a Spec is a nested data-structure consisting of dictionaries and
    sequences, where the leaves are sequences of strings. An example of a Spec is as
    follows:

    >>> spec = Spec({"foo": ["a", "b"], "bar": ["b"]})

    The above ``spec`` describes a dictionary of arrays. As data, it could look
    something like this:

    >>> import numpy as np
    >>> data = {"foo": np.ones([2, 3]), "bar": np.ones([3])}

    Note that the structure of the ``spec`` mirrors the structure of the data, but
    where each array has been replaced with a list of strings, representing the
    dimensions of the arrays. Note also that the second dimension of the ``"foo"``
    array share the same name as the first dimension of the ``"bar"`` array, meaning
    that they are semantically the same dimension. This is better understood with a
    more concrete example:

    >>> spec = Spec({"image":   ["batch", "width", "height", "channels"],
    ...              "caption": ["batch", "tokens"]})

    In the above example, both the ``"image"`` and the ``"caption"`` has the same
    ``"batch"`` dimension so we know that if we loop over the batch-items we must loop
    over both the images and captions.
    """

    def is_leaf(self, tree: Optional[Tree] = "_NOT_GIVEN") -> bool:
        """Return True if this spec object represents the dimensions of an array
        (i.e.: not a nested data-structure of arrays).

        May optionally be called as a static method where ``self`` is a tree, or with
        an explicitly given tree.

        See also:
            func:`._is_spec_leaf`).
        """
        if tree == "_NOT_GIVEN":  # `None` has a semantic meaning
            if isinstance(self, Spec):
                return _is_spec_leaf(self.tree)
            else:
                # is_leaf was called as a static method
                return _is_spec_leaf(self)
        else:
            return _is_spec_leaf(tree)

    def remove_dimension(
        self,
        dimension: Union[str, Sequence[str]],
        path: Sequence = (),
    ) -> "Spec":
        """Remove the given dimension from everywhere in the spec.

        >>> spec = Spec({"signal": ["transmits", "receivers"],
        ...              "receiver": {"position": ["receivers"], "direction": []}})
        >>> spec.remove_dimension("receivers")
        Spec({'signal': ['transmits'], 'receiver': {'position': [], 'direction': []}})

        You can also remove multiple dimensions at once:

        >>> spec.remove_dimension(["transmits", "receivers"])
        Spec({'signal': [], 'receiver': {'position': [], 'direction': []}})
        """
        state = self.get(path)

        if isinstance(dimension, (list, tuple, set)):
            for dim in dimension:
                state = state.remove_dimension(dim, path)
            return state

        for leaf in leaves(state.tree, self.is_leaf):
            if dimension in leaf.value:
                state = state.set([x for x in leaf.value if x != dimension], leaf.path)
        return state

    def index_for(self, dimension: str, path: Sequence = ()) -> Tree:
        """Return the indices of the given dimension in the spec with the same
        structure as the spec.

        >>> spec = Spec({"signal": ["transmits", "receivers"],
        ...              "receiver": {"position": ["receivers"], "direction": []}})
        >>> spec.index_for("receivers")
        {'signal': 1, 'receiver': {'position': 0, 'direction': None}}
        """
        state = self.get(path).tree
        for leaf in leaves(state, self.is_leaf):
            index = (
                leaf.value.index(dimension)
                if (leaf.value is not None and dimension in leaf.value)
                else None
            )
            state = trees.set(state, index, leaf.path)
        return state

    @property
    def dimensions(self) -> Set[str]:
        """Return all dimensions in the spec.

        >>> spec = Spec({"signal": ["transmits", "receivers"],
        ...              "receiver": {"position": ["receivers"], "direction": []},
        ...              "point_position": ["transmits", "points"]})
        >>> sorted(spec.dimensions)
        ['points', 'receivers', 'transmits']
        """
        return reduce(
            lambda dims, leaf: dims.union(leaf.value if leaf.value is not None else []),
            leaves(self.tree, self.is_leaf),
            set(),
        )

    def has_dimension(self, *dimensions: str) -> bool:
        """Return True if the spec has the given dimension(s).

        >>> spec = Spec({"signal": ["transmits", "receivers"],
        ...              "receiver": {"position": ["receivers"], "direction": []}})
        >>> spec.has_dimension("transmits", "receivers")
        True
        >>> spec.has_dimension("frames", "transmits", "receivers")
        False
        """
        return all(dim in self.dimensions for dim in dimensions)

    def add_dimension(
        self, dimension: str, path: Sequence = (), index: int = 0
    ) -> "Spec":
        """Add the dimension to the list of dimensions at the specified path and at the
        specified index in the list.

        >>> spec = Spec({"foo": {"baz": ["a", "b"]}, "bar": ["b"]})
        >>> spec.add_dimension("c", ["foo", "baz"], 0)
        Spec({'foo': {'baz': ['c', 'a', 'b']}, 'bar': ['b']})
        >>> spec.add_dimension("c", ["foo", "baz"], 1)
        Spec({'foo': {'baz': ['a', 'c', 'b']}, 'bar': ['b']})
        >>> spec.add_dimension("c", ["bar"], 0)
        Spec({'foo': {'baz': ['a', 'b']}, 'bar': ['c', 'b']})
        """
        current_dims = self.get(path)
        current_dims = current_dims.tree if current_dims is not None else []
        if not self.is_leaf(current_dims):
            raise ValueError(
                f"The provided path does not lead to a dimensions definition. \
Dimensions must be a list of strings, but got {current_dims} at the path {path}."
            )
        new_dims = [*current_dims[:index], dimension, *current_dims[index:]]
        return self.set(new_dims, path)

    def replace(self, replacements: Tree) -> "Spec":
        """Update the spec by replacing subtrees with corresponding subtrees in the
        replacements tree.

        * A value of None in the replacements tree removes the subtree at the
          corresponding path.
        * A leaf (list of dimensions, see Spec.is_leaf) in the replacements tree always
          replaces the leaf (or subtree) at the corresponding path.
        * Keys present in the replacements tree but not in the spec are added to the
          spec at the corresponding path.

        >>> spec = Spec({"foo": {"baz": ["a", "b"]}, "bar": ["b"]})

        Replacing a path with None removes the subtree at that path:

        >>> spec.replace({"foo": None})
        Spec({'bar': ['b']})

        Removing a subtree such that its parent becomes an empty collection also
        removes the parent:

        >>> spec.replace({"foo": {"baz": None}})
        Spec({'bar': ['b']})

        Replacing an existing path with a list of dimensions overwrites the path:

        >>> spec.replace({"foo": ["c"]})
        Spec({'foo': ['c'], 'bar': ['b']})

        Other than that, it is assumed that the ``replacements`` tree structure mirrors
        the spec structure:

        >>> spec.replace({"foo": {"baz": ["c"]}})
        Spec({'foo': {'baz': ['c']}, 'bar': ['b']})
        """
        state = self.tree
        for replacement in traverse(replacements, self.is_leaf):
            if replacement.value is None:
                state = trees.remove(state, replacement.path)
            elif replacement.is_leaf or self.is_leaf(
                trees.get(state, replacement.path, None)
            ):
                state = trees.set(state, replacement.value, replacement.path)
            else:
                replacement_value = trees.filter(
                    replacement.value,
                    self.is_leaf,
                    lambda tree: tree is not None,
                )
                state = trees.update(
                    state,
                    # Current value takes presedence over replacement value in order to
                    # preserve replace semantics.
                    lambda current_value: trees.merge(
                        replacement_value, current_value, "last"
                    ),
                    replacement.path,
                )
        return Spec(state).prune_empty_branches()

    def validate(self, data: Tree):
        """Validate that the data conforms to the spec, raising a
        :class:`~spekk.validation.ValidationError` if not.

        See also:
            :func:`~spekk.validation.validate`
        """
        from spekk.util.validation import validate

        validate(self, data)

    def size(
        self,
        data: Tree,
        dimension: Optional[str] = None,
    ) -> Union[int, Dict[str, int]]:
        """Get the size of dimensions (or a single dimension) in the data.

        >>> import numpy as np
        >>> spec = Spec({"signal": ["transmits", "receivers"],
        ...              "receiver": {"position": ["receivers"], "direction": []}})
        >>> data = {"signal": np.random.randn(10, 20),
        ...         "receiver": {"position": np.random.randn(20, 3),
        ...                      "direction": np.random.randn(20, 3)}}
        >>> spec.size(data) == {'transmits': 10, 'receivers': 20}
        True
        >>> spec.size(data, "transmits")
        10
        >>> spec.size(data, "receivers")
        20
        """
        from spekk import util

        if dimension is None:
            return {dim: self.size(data, dim) for dim in self.dimensions}
        if not self.has_dimension(dimension):
            raise ValueError(f"Spec does not contain the dimension {dimension}.")

        indices_tree = self.index_for(dimension)
        for leaf in leaves(indices_tree, lambda x: isinstance(x, int) or x is None):
            if leaf.value is not None:
                # Assume that all data with the same dimension has the same size, so we
                # just return the first one we find.
                return util.shape(trees.get(data, leaf.path))[leaf.value]

    def __hash__(self):
        return hash(self.tree)

    def __eq__(self, other) -> bool:
        """Return True if the specs are equal.

        >>> spec = Spec({"signal": ["transmits", "receivers"]})
        >>> spec == Spec({"signal": ["transmits", "receivers"]})
        True
        >>> spec == Spec({"signal": ["frames", "transmits", "receivers"]})
        False
        >>> spec == Spec({"signal": ["transmits", "receivers"], "foo": ["bar"]})
        False
        >>> spec == Spec({"foo": ["bar"]})
        False
        """
        if not isinstance(other, Spec):
            return False
        else:
            for subtree in traverse(other.tree, self.is_leaf):
                if subtree.is_leaf:
                    if not self.has_subtree(subtree.path):
                        return False
                    if len(subtree.value) != len(self.get(subtree.path)):
                        return False
                    for d1, d2 in zip(subtree.value, self.get(subtree.path)):
                        if isinstance(d2, Spec):
                            d2 = d2.tree
                        if d1 != d2:
                            return False
                else:
                    if (
                        treedef(self.get(subtree.path)).keys()
                        != treedef(subtree.value).keys()
                    ):
                        return False
            return True

    def __len__(self):
        return len(self.tree)

    def __repr__(self):
        if self.tree is None:
            return "Spec()"
        return f"Spec({self.tree})"


register_dispatch_fn(lambda t: treedef(t.tree) if isinstance(t, Spec) else None)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



""":func:`flatten` flattens a tree-like data structure into a list of tree-like data 
structures given a dimension and a spec, such that each item in the list either does 
not contain the dimension or it contains *ONLY* the dimension. In effect, we are 
de-structuring the data structure such that it becomes possible to easily loop over the 
dimension for arbitrary data structures."""

from dataclasses import dataclass, field
from typing import Any, Callable, List, Tuple, Union

from spekk import Spec, trees


@dataclass
class _DummyContainer:
    """A temporary container that keeps the structure of the original object and can be
    used to get the original object back given a flattened object.

    All values of data are either another _DummyContainer or an integer representing
    the index in the flattened_obj list of the original value. When calling unflatten,
    the index is looked up in the flattened_obj to get back the original value.
    """

    tree: trees.TreeDef
    data: dict

    def unflatten(self, flattened_obj: list):
        return self.tree.create(
            self.data.keys(),
            [
                v.unflatten(flattened_obj)
                if isinstance(v, _DummyContainer)
                else flattened_obj[v]
                for v in self.data.values()
            ],
        )


@dataclass
class _State:
    """Keep track of the list of parts of the flattened object and their corresponding
    axes (calculated given the spec and a dimension).

    Fields:
    - flattened: the list of all parts of the flattened object.
    - in_axes: the axis of a dimension for each part of the flattened object.
    - i: the current index in flattened. It is incremented every time a part is added
        to flattened (and in_axes).
    """

    flattened: list = field(default_factory=list)
    in_axes: list = field(default_factory=list)
    i: int = 0

    def append(self, value: Any, axis: Union[int, None]) -> int:
        """Add value to flattened and the axis of a dimension for the value to in_axes.

        Return the index of the newly added value and axis in the flattened list.
        """
        self.flattened.append(value)
        self.in_axes.append(axis)
        flattened_index = self.i  # Return the value of i before incrementing
        self.i += 1
        return flattened_index


def flatten(
    obj: trees.Tree,
    spec: Spec,
    dimension: str,
) -> Tuple[List[Any], List[Union[None, int]], Callable[[List[Any]], trees.Tree],]:
    """Flatten/decompose the obj into a flattened list such that parts with the given
    dimension are separate items from those without the dimension.

    This allows us to define objects with fields that may consist of different
    dimensions, but still be able to vmap over them. vmap requires all axes vmapped
    over to have the same size so we need to separate fields with different dimensions
    to different arguments.

    Returns a tuple of:
    - The flattened/decomposed object as a list.
    - The axis of each element in the flattened list that corresponds to the given
        dimension.
    - A function that takes a flattened/decomposed object (same structure as the one
        returned from this function) and returns an object with the same structure as
        the original object.
    """
    state = _State()  # state is updated inside the _flatten function

    def _flatten(obj: trees.Tree, spec: Spec):
        "Recursively flatten obj, mutating the state along the way."

        # Base case 1: the object is not a tree-like structure and can not be
        #   recursively flattened.
        base_case1 = not trees.has_treedef(obj)
        # Base case 2: the spec is a leaf, i.e.: not a tree-like structure, and can not
        #   be used to recursively flatten obj.
        base_case2 = spec.is_leaf()
        if base_case1 or base_case2:
            # Just add the object as-is to the flattened arguments
            flattened_index = state.append(obj, spec.index_for(dimension))
            return flattened_index

        # Else, it is a nested tree-like structure
        tree = trees.treedef(obj)

        # Build up a dummy-object that references the indices in the flattened array
        dummy_data = {}
        for key in tree.keys():
            value = tree.get(key)
            if value is not None:  # None values can be ignored
                sub_spec = spec.get([key]) if spec.has_subtree([key]) else None
                if sub_spec is not None and sub_spec.has_dimension(dimension):
                    dummy_data[key] = _flatten(value, sub_spec)
                else:
                    flattened_index = state.append(value, None)
                    dummy_data[key] = flattened_index
        return _DummyContainer(tree, dummy_data)

    dummy = _flatten(obj, spec)
    if isinstance(dummy, _DummyContainer):
        unflatten = dummy.unflatten
    else:
        # Special case where the object was not a nested structure. If it was just a
        # single item, then we need to just return that item when unflattening.
        unflatten = lambda x: x[0]

    return state.flattened, state.in_axes, unflatten



import numpy as np

from spekk import Spec, trees


def generate_mock_data(
    spec: Spec,
    seed: int = 0,
    min_dimension_size=1,
    max_dimension_size=4,
):
    rng = np.random.default_rng(seed)
    dim_sizes = {
        dim: rng.integers(min_dimension_size, max_dimension_size)
        for dim in spec.dimensions
    }
    return trees.update_leaves(
        spec.tree,
        spec.is_leaf,
        lambda dims: rng.random([dim_sizes[dim] for dim in dims]),
    )



"""Module for abstracting over tree-like data structures; in essence, everything that 
is tree-like can be represented as a mapping of keys and values."""

from abc import ABC, abstractmethod
from typing import Any, Callable, Mapping, Sequence, Union

Tree = Union[Mapping[Any, "Tree"], Sequence["Tree"], Any]


class TreeDef(ABC):
    """A :class:`TreeDef` is an abstraction for a tree-like data structure.

    We must be able to get the subtrees of the tree (keys and get) and create a copy
    with updated values (create). Anything that can support these operations can be used
    as a tree.
    """

    def __init__(self, tree: Tree):
        self.tree = tree

    @abstractmethod
    def keys(self) -> Sequence:
        "Get the keys that can be used to get each subtree."

    @abstractmethod
    def get(self, key: Any):
        "Get the subtree at the given key."

    def values(self) -> Sequence:
        "Get the subtrees for each key in the treedef."
        return [self.get(k) for k in self.keys()]

    @abstractmethod
    def create(self, keys: Sequence, values: Sequence) -> Tree:
        """Create a new instance of the tree with the given keys and values.

        >>> td = treedef({"a": 1, "b": 2})
        >>> td.create(["a", "b"], [3, 4])
        {'a': 3, 'b': 4}
        """

    @staticmethod
    def new_class(
        keys_fn: Callable[[Tree], Sequence],
        get_fn: Callable[[Tree, Any], Any],
        create_fn: Callable[[Sequence, Sequence], Tree],
    ) -> "TreeDef":
        "Helper function for creating a new :class:`TreeDef` class."

        class _TreeDef(TreeDef):
            def keys(self) -> Sequence:
                return keys_fn(self.tree)

            def get(self, key: Any):
                return get_fn(self.tree, key)

            def create(self, keys: Sequence, values: Sequence) -> Tree:
                return create_fn(keys, values)

        return _TreeDef
    
    def items(self):
        return zip(self.keys(), self.values())


class DuckTypedTreeDef(TreeDef):
    """An object can be a :class:`TreeDef` if it has the dunder-methods: 
    ``__spekk_treedef_keys__``, ``__spekk_treedef_get__``, and 
    ``__spekk_treedef_create__``."""
    def __init__(self, obj: Any):
        if not (
            hasattr(obj, "__spekk_treedef_keys__")
            and hasattr(obj, "__spekk_treedef_get__")
            and hasattr(obj, "__spekk_treedef_create__")
        ):
            raise ValueError(
                f"Object {obj} does not have the required dunder-methods to be a treedef."
            )
        self.obj = obj

    def keys(self) -> Sequence:
        return getattr(self.obj, "__spekk_treedef_keys__")()

    def get(self, key: Any):
        return getattr(self.obj, "__spekk_treedef_get__")(key)

    def create(self, keys: Sequence, values: Sequence) -> Any:
        return getattr(self.obj, "__spekk_treedef_create__")(keys, values)


# A registry of types to TreeDef's.
type_registry = {}


def dispatch_treedef(tree: Tree):
    "If the tree itself is a :class:`TreeDef`, return it."
    if isinstance(tree, TreeDef):
        return tree


def dispatch_by_type(tree: Tree):
    """Given a tree, return the :class:`TreeDef` for its type (through the 
    ``type_registry``)."""
    t = type(tree)
    if t in type_registry:
        return type_registry[type(tree)](tree)


def dispatch_by_duck_type(tree: Tree):
    """Given a tree, return a :class:`TreeDef` if it has the required dunder-methods. 
    See :class:`DuckTypedTreeDef` for more details."""
    try:
        return DuckTypedTreeDef(tree)
    except ValueError:
        return None


# A registry of functions that can be used to get a TreeDef for a given tree (just a
# list of functions that are tried in order until one of them returns a TreeDef).
dispatch_fn_registry = [dispatch_treedef, dispatch_by_type, dispatch_by_duck_type]


def register_type(t: type, treedef: TreeDef):
    "Register a new :class:`TreeDef` by type."
    type_registry[t] = treedef


def register_dispatch_fn(dispatch_fn: Callable[[Tree], Union[TreeDef, None]]):
    """Register a new :class:`TreeDef` by dispatch function.

    Multiple dispatch functions may be registered, in which case they will be tried in
    order until one of them returns a :class:`TreeDef`. Note that 
    :func:`dispatch_by_type` and :func:`dispatch_treedef` always comes first and takes 
    precendence."""
    dispatch_fn_registry.append(dispatch_fn)


def treedef(tree: Tree) -> TreeDef:
    """Return the :class:`TreeDef` (if registered) for the given tree (``dict``, 
    ``list``, and ``tuple`` are registered by default)."""
    for dispatch_fn in dispatch_fn_registry:
        td = dispatch_fn(tree)
        if td:
            return td
    raise ValueError(
        f"No TreeDef found for {repr(tree)}. Perhaps you need to register one?"
    )


def has_treedef(tree: Tree) -> bool:
    """Return ``True`` if a :class:`TreeDef` is registered for the given tree."""
    try:
        treedef(tree)
        return True
    except ValueError:
        return False


# Register some basic tree types
register_type(
    dict,
    TreeDef.new_class(
        lambda d: d.keys(),
        lambda d, k: d[k],
        lambda keys, values: {k: v for k, v in zip(keys, values)},
    ),
)
register_type(
    list,
    TreeDef.new_class(
        lambda l: range(len(l)),
        lambda l, i: l[i],
        lambda keys, values: list(values),
    ),
)
register_type(
    tuple,
    TreeDef.new_class(
        lambda t: range(len(t)),
        lambda t, i: t[i],
        lambda keys, values: tuple(values),
    ),
)

if __name__ == "__main__":
    import doctest

    doctest.testmod()



":class:`Apply` applies function ``f`` to the output of the wrapped function."

from typing import Callable

import spekk.transformations.common as common
from spekk import Spec, trees
from spekk.transformations.axis import Axis, concretize_axes
from spekk.transformations.base import Transformation


class Apply(Transformation):
    """Transform a function such that ``f`` is applied to the output of it.
    
    Attributes:
        f: The function to apply to the result of the wrapped function.
        args: Optional extra positional arguments to pass to ``f``.
        kwargs: Optional extra keyword arguments to pass to ``f``.
    """
    def __init__(self, f: Callable, *args, **kwargs):
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def transform_function(
        self, to_be_transformed: callable, input_spec: Spec, output_spec: Spec
    ) -> callable:
        def with_applied_f(*args, **kwargs):
            result = to_be_transformed(*args, **kwargs)
            args, kwargs = concretize_axes(output_spec, self.args, self.kwargs)
            return self.f(result, *args, **kwargs)

        return with_applied_f

    def transform_input_spec(self, spec: Spec) -> Spec:
        return spec

    def transform_output_spec(self, spec: Spec) -> Spec:
        tree = (self.args, self.kwargs)
        for leaf in trees.leaves(
            tree, lambda x: isinstance(x, Axis) or not trees.has_treedef(x)
        ):
            if isinstance(leaf.value, Axis):
                spec = spec.update_leaves(leaf.value.new_dimensions)
        return spec

    def __repr__(self) -> str:
        args_str = ", ".join([str(arg) for arg in self.args])
        kwargs_str = ", ".join([f"{k}={str(v)}" for k, v in self.kwargs.items()])
        repr_str = f"Apply({common.get_fn_name(self.f)}"
        if self.args:
            repr_str += f", {args_str}"
        if self.kwargs:
            repr_str += f", {kwargs_str}"
        # Make sure the repr string is not too long
        if len(repr_str) > 140:
            repr_str = repr_str[: (140 - len("… <truncated>"))] + "… <truncated>"
        return repr_str + ")"



import jax
import numpy as np

from spekk import Spec, util
from spekk.transformations import ForAll, compose


def test_multiple_for_alls():
    f = lambda x: x + 1
    data = {"x": np.ones((2, 3, 4, 5))}
    spec = Spec({"x": ["2", "3", "4", "5"]})

    tf1 = compose(f, ForAll("3"), ForAll("5"), ForAll("2"), ForAll("4")).build(spec)
    tf2 = compose(f, ForAll("4", "2", "5", "3")).build(spec)  # Reversed order

    assert tf1(**data) == tf2(**data)
    assert tf1.input_spec == tf2.input_spec
    assert tf1.output_spec == tf2.output_spec



"Functions that operate on trees."

import operator
from dataclasses import dataclass
from typing import Any, Callable, Generator, Sequence

from spekk.trees.registry import Tree, treedef

_NO_DEFAULT = object()


def update(tree: Tree, f: Callable[[Tree], Tree], path: tuple):
    """Update the subtree at the given path.

    >>> tree = {"a": [1, {"b": 2}, 3], "c": 4}
    >>> update(tree, lambda x: x + 10, ("a", 1, "b"))
    {'a': [1, {'b': 12}, 3], 'c': 4}
    """
    if not path:
        return f(tree)

    key, *remaining_path = path
    td = treedef(tree)
    values = [
        update(td.get(k), f, remaining_path) if k == key else td.get(k)
        for k in td.keys()
    ]
    return td.create(td.keys(), values)


def get(tree: Tree, path: tuple, default: Any = _NO_DEFAULT):
    """Get the subtree at the given path.

    >>> tree = {"a": [1, {"b": 2}, 3], "c": 4}
    >>> get(tree, ("a", 1, "b"))
    2
    """
    if not path:
        return tree

    key, *remaining_path = path
    td = treedef(tree)
    if default is not _NO_DEFAULT:
        if key not in td.keys():
            return default
    return get(td.get(key), remaining_path)


def set(tree: Tree, value: Any, path: tuple):
    """Set the value of the subtree at the given path.

    >>> tree = {"a": [1, {"b": 2}, 3], "c": 4}
    >>> set(tree, 42, ("a", 1, "b"))
    {'a': [1, {'b': 42}, 3], 'c': 4}
    """
    return update(tree, lambda _: value, path)


def remove(tree: Tree, path: tuple):
    """Remove the value of the subtree at the given path.

    >>> tree = {"a": [1, {"b": 2}, 3], "c": 4, "d": 5}
    >>> remove(tree, ("a", 1, "b"))
    {'a': [1, {}, 3], 'c': 4, 'd': 5}
    >>> remove(tree, ("a", 1))
    {'a': [1, 3], 'c': 4, 'd': 5}
    >>> remove(tree, ("c",))
    {'a': [1, {'b': 2}, 3], 'd': 5}
    """

    def remove_sub_tree(tree):
        td = treedef(tree)
        keys = [k for k in td.keys() if k != path[-1]]
        values = [td.get(k) for k in keys]
        return td.create(keys, values)

    return update(tree, remove_sub_tree, path[:-1])


def merge(t1: Tree, t2: Tree, preserve_order: int = "first") -> Tree:
    """Merge two trees (assuming this is possible).

    The order of the keys in the merged tree is determined by the preserve_order. If it
    is first, then the order is the same as the first tree. If it is last, then the
    order is the same as the second tree. Any other new keys are appended to the end of
    the tree.

    >>> merge({"a": 1, "b": 2}, {"c": 3})
    {'a': 1, 'b': 2, 'c': 3}
    >>> merge({"a": 1, "b": 2}, {"b": 7, "c": 3})
    {'a': 1, 'b': 7, 'c': 3}

    Merging two lists treats the indices as keys, and the second tree overwrites the
    indices of the first one.

    >>> merge([1, 2, 3], [4, 5])
    [4, 5, 3]
    """
    treedef1, treedef2 = treedef(t1), treedef(t2)

    if preserve_order == "first":
        merged = dict(zip(treedef1.keys(), treedef1.values()))
        for key in treedef2.keys():
            merged[key] = treedef2.get(key)
    elif preserve_order == "last":
        merged = dict(zip(treedef2.keys(), treedef2.values()))
        for key in treedef1.keys():
            if key not in merged:
                merged[key] = treedef1.get(key)
    else:
        raise ValueError("preserve_order must be either 'first' or 'last'")

    return treedef1.create(merged.keys(), merged.values())


@dataclass
class TraversalItem:
    "An object returned from the :func:`traverse` and :func:`leaves` generator functions."
    value: Any  #: The value of the node that was traversed onto.
    path: tuple  #: The path to the node.
    is_leaf: bool  #: Whether the node is a leaf or not.


def traverse(
    tree: Tree,
    is_leaf: Callable[[Tree], bool],
    path: tuple = (),
) -> Generator[TraversalItem, None, None]:
    """Traverse a tree and yield all nodes (subtrees and leaves) as
    :class:`TraversalItem` objects. The traversal is depth-first, left-to-right.

    >>> tree = {"a": [1, {"b": 2}, 3], "c": 4}
    >>> for item in traverse(tree, is_leaf=lambda x: isinstance(x, int)):
    ...     print(item.value)
    1
    2
    {'b': 2}
    3
    [1, {'b': 2}, 3]
    4
    {'a': [1, {'b': 2}, 3], 'c': 4}
    """
    if is_leaf(tree):
        yield TraversalItem(tree, path, True)
    else:
        td = treedef(tree)
        values = []
        for key in td.keys():
            subtrees = list(traverse(td.get(key), is_leaf, path + (key,)))
            for subtree in subtrees:
                yield subtree
            values.append(subtrees[-1].value)
        yield TraversalItem(td.create(td.keys(), values), path, False)


def leaves(
    tree: Tree,
    is_leaf: Callable[[Tree], bool],
    path: tuple = (),
) -> Generator[TraversalItem, None, None]:
    """Traverse a tree and yield all leaves as :class:`TraversalItem` objects. The
    traversal is depth-first, left-to-right.

    >>> tree = {"a": [1, {"b": 2}, 3], "c": 4}
    >>> for item in leaves(tree, is_leaf=lambda x: isinstance(x, int)):
    ...     print(item.value)
    1
    2
    3
    4
    """
    return (t for t in traverse(tree, is_leaf, path) if t.is_leaf)


def filter(
    tree: Tree,
    is_leaf: Callable[[Tree], bool],
    predicate: Callable[[Tree], bool],
    path: tuple = (),
) -> Tree:
    """Remove all subtrees for which the predicate returns ``False``.

    >>> tree = {"a": [1, {"b": 2}, 3], "c": 4, "d": 5}
    >>> is_leaf = lambda x: isinstance(x, int)
    >>> filter(tree, is_leaf, lambda x: x % 2 == 1 if is_leaf(x) else True)
    {'a': [1, {}, 3], 'd': 5}
    """
    state = tree
    for t in traverse(tree, is_leaf, path):
        if not predicate(t.value):
            state = remove(state, t.path)
    return state


def update_leaves(
    tree: Tree,
    is_leaf: Callable[[Tree], bool],
    f: Callable[[Sequence[str]], Sequence[str]],
    path: tuple = (),
) -> Tree:
    """Apply ``f`` to all leaves in the tree.

    >>> tree = {"foo": ["a", "b"], "bar": ["c"]}
    >>> update_leaves(
    ...     tree,
    ...     lambda t: isinstance(t, list),
    ...     lambda dims: dims + ["new_dim"])
    {'foo': ['a', 'b', 'new_dim'], 'bar': ['c', 'new_dim']}
    """
    state = tree
    for leaf in leaves(tree, is_leaf, path):
        state = set(state, f(leaf.value), leaf.path)
    return state


def are_equal(
    tree1: Tree,
    tree2: Tree,
    is_leaf: Callable[[Tree], bool],
    leafs_are_equal: Callable[[Any, Any], bool] = operator.eq,
    path: tuple = (),
) -> bool:
    """Return ``True`` if the two trees has the same structure and each leaf are equal
    according to ``leafs_are_equal`` (defaults to ``==`` operator)."""
    for t in traverse(tree1, is_leaf, path):
        # Check if the path exists in the other tree.
        try:
            other_value = get(tree2, t.path)
        except KeyError:
            return False

        if t.is_leaf:
            if not leafs_are_equal(t.value, other_value):
                return False
        else:
            treedef1 = treedef(t.value)
            treedef2 = treedef(other_value)
            if tuple(treedef1.keys()) != tuple(treedef2.keys()):
                return False
    return True


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import numpy as np

from spekk import Spec
from spekk.transformations import ForAll, compose

kernel = lambda x: x**2
data = {"x": np.ones((2, 3)) * 2}
spec = Spec({"x": ["b", "a"]})

forall_xy = compose(ForAll("a"), ForAll("b"))
tf_partial = compose(kernel, forall_xy).build(spec)
tf_full = compose(kernel, ForAll("a"), ForAll("b")).build(spec)

np.testing.assert_equal(tf_partial(**data), tf_full(**data))
print(np.array_equal(tf_partial(**data), tf_full(**data)))


from random import Random

import hypothesis as ht
import hypothesis.strategies as st
import numpy as np
from test_helpers.generators.spec import specs
from test_helpers.mock_data import generate_mock_data

from spekk import Spec, trees, util


def test_flattening_dict():
    foo_x = np.ones((2, 3)) * 4
    foo_y = np.ones((2,)) * 2
    bar = np.ones((3,)) * 3
    baz_z = np.ones((4,))
    obj = {
        "foo": {
            "x": foo_x,
            "y": [foo_y, foo_y],
        },
        "bar": bar,
        "baz": {"z": baz_z},
    }
    spec = Spec(
        {
            "foo": {
                "x": ["b", "a"],
                "y": [["a"], ["a"]],
            }
        }
    )

    # "x" and "y" should be split apart since "x" contains dimension "b"
    flattened, in_axes, unflatten = util.flatten(obj, spec, "b")
    np.testing.assert_equal(flattened, [foo_x, [foo_y, foo_y], bar, {"z": baz_z}])
    assert in_axes == [0, None, None, None]
    np.testing.assert_equal(obj, unflatten(flattened))

    # "x" and the list items under "y" should be split apart since they all contain
    # dimension "a"
    flattened, in_axes, unflatten = util.flatten(obj, spec, "a")
    np.testing.assert_equal(flattened, [foo_x, foo_y, foo_y, bar, {"z": baz_z}])
    assert in_axes == [1, 0, 0, None, None]
    np.testing.assert_equal(obj, unflatten(flattened))


@ht.given(specs())
def test_flatten_unflatten_data(spec: Spec):
    """Test that for any spec with corresponding valid data, flattening and 
    unflattening the data for any dimension results in the original data."""
    data = generate_mock_data(spec)
    for dimension in spec.dimensions:
        flattened, _, unflatten = util.flatten(data, spec, dimension)
        assert trees.are_equal(
            unflatten(flattened),
            data,
            lambda x: isinstance(x, (np.ndarray, float, int)),
            np.array_equal,
        )

@ht.given(specs())
def test_flattened_in_axes(spec: Spec):
    """Test that for any spec with corresponding valid data, the `in_axes` calculated 
    from flattening the data have at least 1 non-None value."""
    data = generate_mock_data(spec)
    for dimension in spec.dimensions:
        _, in_axes, _ = util.flatten(data, spec, dimension)
        assert not all(x is None for x in in_axes)



"Generalized :func:`np.shape` that works on nested Python sequences."

from typing import Sequence


def shape(x) -> Sequence[int]:
    """Get the shape of an array, number, or a nested sequence of numbers.

    >>> shape(1.0)
    ()
    >>> shape([0, 1, 2])
    (3,)
    >>> shape([[0, 1, 2], [3, 4, 5]])
    (2, 3)

    >>> import numpy as np
    >>> shape(np.ones((2, 3)))
    (2, 3)
    """
    if isinstance(x, (int, float, complex)):
        return ()
    elif hasattr(x, "shape"):
        return x.shape
    elif isinstance(x, (list, tuple, range)):
        # Assume each item in x has the same shape.
        return (len(x), *shape(x[0]))
    raise ValueError(f"Cannot get shape of object with type {type(x)}")


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"Validate data according to a :class:`~spekk.spec.Spec`."

from dataclasses import dataclass
from typing import Generator, Sequence

from spekk import trees, util
from spekk.spec import Spec


class ValidationError(Exception):
    "Raised when the data does not conform to the spec."


@dataclass
class _DimensionSizeInfo:
    "Helper class for keeping track of the shape of data for a dimension."
    shape: Sequence[int]  # The shape of the value
    index: int  # The index in the shape for the dimension
    path: tuple  # The path to the value

    @property
    def size(self):
        "The size of the dimension in the data."
        return self.shape[self.index]


def _paths_and_indices(spec: Spec, dimension: str) -> Generator[tuple, None, None]:
    """Yield paths and indices for the given dimension.

    For example, given a spec as such:

    >>> spec = Spec({
    ...     "k1": {"k1_sub": ["dim1", "dim2"]},
    ...     "k2": ["dim2"]
    ... })

    This is what we get if we print out each path and index for the dimension "dim1":

    >>> for path, index in list(_paths_and_indices(spec, "dim1")):
    ...     print(path, index)
    ('k1', 'k1_sub') 0

    For dimension "dim2", both the keys "k1" and "k2" contain the dimension:

    >>> for path, index in list(_paths_and_indices(spec, "dim2")):
    ...     print(path, index)
    ('k1', 'k1_sub') 1
    ('k2',) 0
    """
    # An index can either be int or None
    is_leaf = lambda x: isinstance(x, int) or x is None
    indices = spec.index_for(dimension)
    for subtree in trees.leaves(indices, is_leaf):
        index = subtree.value
        if index is not None:
            yield subtree.path, index


def _check_path_present_in_data(data, path):
    """Return the value at the given path in the data, raising a
    :class:`ValidationError` if it is not present."""
    try:
        return trees.get(data, path)
    except KeyError:
        raise ValidationError(
            f"Path {list(path)} is not present in the data yet it is present in the spec."
        ) from None


def _check_value_has_shape_attribute(value, path):
    """Return the shape of the value, raising a :class:`ValidationError` if it does not
    have a ``__spekk_shape__`` or ``shape`` attribute."""
    try:
        return util.shape(value)
    except Exception as e:
        raise ValidationError(
            f"Unable to get the shape of value at path {list(path)} with type {type(value)}."
        ) from e


def _check_value_has_dimensions(spec, shape, path):
    """Raise a :class:`ValidationError` if the value has fewer dimensions than the spec.

    If the spec has more dimensions than the data then the data is not valid for the
    spec. The data may however have more dimensions than the spec, for example if the
    data has any additional dimensions that we don't care about, like the
    ``xyz``-dimension of a point coordinate. The additional dimensions are assumed to
    come last in the shape."""
    spec_dimensions_at_path = trees.get(spec, path)
    if len(shape) < len(spec_dimensions_at_path):
        raise ValidationError(
            f"The data has only {len(shape)} dimension{'s' if len(shape)>1 else ''} \
while the spec specifies dimensions {spec_dimensions_at_path} \
({len(spec_dimensions_at_path)} in total) at the path {list(path)}."
        )


def _check_consistent_dimension_sizes(
    dimension: str,
    dimension_sizes_info: Sequence[_DimensionSizeInfo],
):
    """Raise a :class:`ValidationError` if the dimension has inconsistent sizes in the
    data.

    This can happen if the data has arrays with different sizes for a given dimension.
    All array-sizes corresponding to the shape dimension must have the same size."""
    dimension_sizes = [info.size for info in dimension_sizes_info]
    if len(set(dimension_sizes)) > 1:  # If there are more than 1 distinct value
        path_sizes_str = "\n".join(
            f"    - Size={info.size} at path {list(info.path)}, shape={info.shape}, index={info.index}"
            for info in dimension_sizes_info
        )
        raise ValidationError(
            f"Dimension '{dimension}' has inconsistent sizes in the data:\n{path_sizes_str}"
        )


def validate(spec: Spec, data: trees.Tree):
    """Validate that the data conforms to the spec, raising a :class:`ValidationError`
    if not.

    Examples:

    >>> import numpy as np
    >>> spec = Spec({
    ...     "foo": {"bar": ["dim1", "dim2"]},
    ...     "baz": ["dim2"],
    ... })

    The following data is valid because ``"dim2"`` under path ``["foo", "bar"]`` has
    the same size as under path ``["baz"]`` and it has the same structure as the spec:

    >>> validate(spec, {
    ...     "foo": {"bar": np.ones((2, 3))},
    ...     "baz": np.ones((3,)),
    ... })

    If we try to validate against a spec with a path that is not present in the data, a
    :class:`ValidationError` is raised:

    >>> validate(spec, {
    ...     "foo": {"invalid_key": np.ones((2, 3))},
    ...     "baz": np.ones((3,)),
    ... })
    Traceback (most recent call last):
        ...
    ValidationError: Path ['foo', 'bar'] is not present in the data yet it is present in the spec.

    All specced values must have a ``shape`` attribute (or more generally; be supported
    by :func:`spekk.util.shape.shape`):

    >>> validate(spec, {
    ...     "foo": {"bar": np.ones((2, 3))},
    ...     "baz": object(),  # <- An object does not have a shape attribute
    ... })
    Traceback (most recent call last):
        ...
    ValidationError: Unable to get the shape of value at path ['baz'] with type <class 'object'>.

    The data must have at least as many dimensions as the spec:

    >>> validate(
    ...     Spec({"foo": ["dim1", "dim2"]}),
    ...     {"foo": np.ones((2,))},  # <- Too few dimensions!
    ... )
    Traceback (most recent call last):
        ...
    ValidationError: The data has only 1 dimension while the spec specifies dimensions ['dim1', 'dim2'] (2 in total) at the path ['foo'].

    It is OK if the data has more dimensions than the spec:

    >>> validate(
    ...     Spec({"foo": ["dim1", "dim2"]}),    # <- spec specifies 2 dimensions
    ...     {"foo": np.ones((2, 3, 4, 5, 6))},  # <- 5 dimensions is more than 2, which is OK!
    ... )

    The data must have the same size for a given dimension in all places where it is
    used:

    >>> validate(spec, {
    ...     "foo": {"bar": np.ones((2, 3))},
    ...     "baz": np.ones((4,)),  # <- Size of dim2 is 4 here, but 3 above!
    ... })
    Traceback (most recent call last):
        ...
    ValidationError: Dimension 'dim2' has inconsistent sizes in the data:
        - Size=3 at path ['foo', 'bar'], shape=(2, 3), index=1
        - Size=4 at path ['baz'], shape=(4,), index=0

    >>> validate(spec, {
    ...     "foo": {"bar": np.ones((2, 3))},
    ...     "baz": np.ones((3,)),  # <- This has the same size as the one above
    ... })

    The spec does not have to specify the dimensions for all paths in the data:

    >>> validate(spec, {
    ...     "foo": {"bar": np.ones((2, 3))},
    ...     "baz": np.ones((3,)),
    ...     "qux": np.ones((5, 6)),  # <- This is OK, the spec does not specify dimensions for "qux"
    ... })
    """
    for dimension in spec.dimensions:
        # A list is used to store information about the shape of the data for this
        # dimension, so we can check for consistency later.
        dimension_sizes_info = []
        for path, index in _paths_and_indices(spec, dimension):
            value = _check_path_present_in_data(data, path)
            shape = _check_value_has_shape_attribute(value, path)
            _check_value_has_dimensions(spec, shape, path)
            # Add the information to a list so we can check for consistency.
            dimension_sizes_info.append(_DimensionSizeInfo(shape, index, path))
        _check_consistent_dimension_sizes(dimension, dimension_sizes_info)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""Classes and functions for working with tree-like structures."""

from spekk.trees.core import (
    are_equal,
    get,
    leaves,
    merge,
    remove,
    set,
    traverse,
    update,
    update_leaves,
)
from spekk.trees.registry import (
    Tree,
    TreeDef,
    has_treedef,
    register_dispatch_fn,
    register_type,
    treedef,
)
from spekk.trees.treelens import TreeLens

__all__ = [
    "are_equal",
    "get",
    "leaves",
    "merge",
    "remove",
    "set",
    "traverse",
    "update",
    "update_leaves",
    "Tree",
    "TreeDef",
    "has_treedef",
    "register_dispatch_fn",
    "register_type",
    "treedef",
    "TreeLens",
]



""":class:`Reduce` transforms a function that works on scalar inputs such that it works 
on arrays instead and iteratively reduces the values of a dimension. For example, 
``Reduce.Sum("dimension")`` produces equivalent results as a ``ForAll("dimension")`` 
followed by an ``Apply(np.sum, "dimension")`` transformation, but will potentially use 
less memory because it sums the partial results iteratively isntead of trying to 
parallelize over the dimension first (in the case of using GPU-backends such as JAX).
"""
import operator
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterable, Optional, Sequence, TypeVar

from spekk import Spec, trees, util
from spekk.transformations import Transformation, common
from spekk.transformations.axis import Axis, concretize_axes

T_reduce_cls = TypeVar("T_reduce_cls", bound="Reduce")
T_f_result = TypeVar("T_f_result")
T_reduction_result = TypeVar("T_reduction_result")
T_reduce_fn = Callable[[T_reduction_result, T_f_result], T_reduction_result]
T_reduce = Callable[[T_reduce_fn, Iterable, T_reduction_result], T_reduction_result]


@dataclass
class Reduce(Transformation):
    """Transform a function to make it reduce the values of a dimension iteratively.

    A :class:`Reduce` transformation is generally a :class:`ForAll` and :class:`Apply`
    transformation combined, if the :class:`Apply` transformation somehow aggregates
    the result (for example by summing over the vectorized axis).

    As a concrete example:
    ``ForAll("transmits")`` followed by ``Apply(np.sum, Axis("transmits")`` is
    equivalent to ``Reduce.Sum("transmits")``, but using ``Reduce.Sum`` will likely
    allocate a lot less memory, potentially at the cost of processing time.
    """

    dimension: str  #: The dimension to reduce over.
    reduce_fn: T_reduce_fn  #: The function to reduce with. For example operator.add for summation.
    initial_value: Optional[
        T_reduction_result
    ] = None  #: The initial value for the reduction. For example 0 for summation.
    enumerate: bool = False  #: If True, each item is a tuple of (index, value), similar to Python's built-in :func:`enumerate`.
    extra_args: Sequence[Any] = field(
        default_factory=tuple
    )  #: Extra arguments to pass to the reduce function.
    extra_kwargs: Dict[str, Any] = field(
        default_factory=dict
    )  #: Extra keyword arguments to pass to the reduce function.
    reduce_impl: Optional[
        T_reduce
    ] = None  #: The ``reduce`` implementation to use. Defaults to Python's built-in :func:`functools.reduce`.

    def __post_init__(self):
        "Sub-classes may override this method to perform additional initialization."

    def transform_function(
        self, to_be_transformed: callable, input_spec: Spec, output_spec: Spec
    ) -> callable:
        if not input_spec.has_dimension(self.dimension):
            raise ValueError(f"Spec does not contain the dimension {self.dimension}.")

        def wrapped(*_unsupported_positional_args, **kwargs):
            if _unsupported_positional_args:
                raise ValueError(
                    "Positional arguments are not supported when using Reduce. Use keyword arguments when calling the transformed function instead."
                )
            extra_args, extra_kwargs = concretize_axes(
                output_spec, self.extra_args, self.extra_kwargs
            )
            reduce_fn = lambda *args, **kwargs: self.reduce_fn(
                *args, *extra_args, **kwargs, **extra_kwargs
            )
            return specced_map_reduce(
                to_be_transformed,
                reduce_fn,
                kwargs,
                input_spec,
                self.dimension,
                self.initial_value,
                self.enumerate,
                self.reduce_impl,
            )

        return wrapped

    def transform_input_spec(self, spec: Spec) -> Spec:
        return spec.remove_dimension(self.dimension)

    def transform_output_spec(self, spec: Spec) -> Spec:
        tree = (self.extra_args, self.extra_kwargs)
        for leaf in trees.leaves(
            tree, lambda x: isinstance(x, Axis) or not trees.has_treedef(x)
        ):
            if isinstance(leaf.value, Axis):
                spec = spec.update_leaves(leaf.value.new_dimensions)
        return spec

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            + f'"{self.dimension}", '
            + f"{self.reduce_fn}, "
            + f"{self.initial_value})"
        )

    @classmethod
    def Sum(
        cls: T_reduce_cls,
        dimension: str,
        initial_value: Optional[T_reduction_result] = None,
        reduce_impl: T_reduce = None,
    ) -> T_reduce_cls:
        """Transformation that iteratively adds the results of the wrapped function for
        each item in the given dimension."""
        return cls(dimension, operator.add, initial_value, reduce_impl=reduce_impl)

    @classmethod
    def Product(
        cls: T_reduce_cls,
        dimension: str,
        initial_value: Optional[T_reduction_result] = None,
        reduce_impl: T_reduce = None,
    ) -> T_reduce_cls:
        """Transformation that iteratively multiplies the results of the wrapped
        function for each item in the given dimension."""
        return cls(dimension, operator.mul, initial_value, reduce_impl=reduce_impl)


def specced_map_reduce(
    map_f: Callable[[Any], Any],
    reduce_f: T_reduce_fn,
    data: Any,
    spec: Spec,
    dimension: str,
    initial_value: Optional[Any] = None,
    enumerate: bool = False,
    reduce_impl: Optional[T_reduce] = None,
):
    # Flatten the data so that we can iterate over it without worrying about the
    # potentially nested structure of the data.
    flattened_args, in_axes, unflatten = util.flatten(data, spec, dimension)

    # Get the size of the dimension that we are reducing over.
    size = spec.size(data, dimension)
    if size == 0:  # Return early if there are no elements to reduce over.
        return initial_value

    # Get the first mapped value.
    carry = common.map_1_flattened(map_f, flattened_args, in_axes, unflatten, 0)
    # Use the first mapped value as the initial value if no initial value was given.
    if initial_value is not None:
        x = (0, carry) if enumerate else carry
        carry = reduce_f(initial_value, x)

    # `wrapped` puts everything together and makes it work with `reduce_impl`.
    # It gets the arguments indexed at `i` for the given dimension, and applies the
    # mapping function to them before performing a reduction step.
    def wrapped(carry, i):
        x = common.map_1_flattened(map_f, flattened_args, in_axes, unflatten, i)
        x = (i, x) if enumerate else x
        return reduce_f(carry, x)

    # Default `reduce_impl` is Python's built-in `functools.reduce`.
    if reduce_impl is None:
        import functools

        reduce_impl = functools.reduce
    # Reduce over the remaining elements. Note that `carry` is the first element, so we
    # start at index 1 in the `range(1, size)`.
    return reduce_impl(wrapped, range(1, size), carry)



import numpy as np

from spekk import Spec

spec = Spec({"foo": ["a", "b"], "bar": ["b"]})
deeper_spec = Spec({"foo": {"baz": ["a", "b"], "quaz": ["c"]}, "bar": ["b"]})
sequence_spec = Spec([["a", "b"], ["b"]])


def test_getitem():
    # Getting the value or subtree at the given path returns a new Spec object.
    assert spec.get(["foo"]) == Spec(["a", "b"])
    assert spec.get(["bar"]) == Spec(["b"])

    assert spec["foo"] == Spec(["a", "b"])
    assert spec["bar"] == Spec(["b"])

    assert spec["foo"].tree == ["a", "b"]
    assert spec["bar"].tree == ["b"]


def test_is_leaf():
    assert not spec.is_leaf()
    assert spec["foo"].is_leaf()
    assert spec["bar"].is_leaf()


def test_remove_dimension():
    assert spec.remove_dimension("a") == Spec({"foo": ["b"], "bar": ["b"]})
    assert spec.remove_dimension("b") == Spec({"foo": ["a"], "bar": []})


def test_index_for():
    assert spec.index_for("a") == {"foo": 0, "bar": None}
    assert spec.index_for("b") == {"foo": 1, "bar": 0}


def test_get_dimensions():
    assert spec.dimensions == {"a", "b"}
    assert spec.remove_dimension("b").dimensions == {"a"}


def test_has_dimension():
    assert spec.has_dimension("a")
    assert spec.has_dimension("b")
    assert not spec.has_dimension("c")


def test_add_dimension():
    assert spec.add_dimension("c", ["foo"], 0) == Spec(
        {"foo": ["c", "a", "b"], "bar": ["b"]}
    )
    assert spec.add_dimension("c", ["foo"], 1) == Spec(
        {"foo": ["a", "c", "b"], "bar": ["b"]}
    )
    assert spec.add_dimension("c", ["foo"], 2) == Spec(
        {"foo": ["a", "b", "c"], "bar": ["b"]}
    )
    assert spec.add_dimension("c", ["bar"], 0) == Spec(
        {"foo": ["a", "b"], "bar": ["c", "b"]}
    )
    assert spec.add_dimension("c", ["bar"], 1) == Spec(
        {"foo": ["a", "b"], "bar": ["b", "c"]}
    )


def test_replace():
    # Removing a path by setting it to None
    assert spec.replace({"foo": None}) == Spec({"bar": ["b"]})
    # Replacing a path with a subtree
    assert spec.replace({"foo": {"baz": []}}) == Spec(
        {"foo": {"baz": []}, "bar": ["b"]}
    )

    # Removing a deeper path with a subtree potentially removes the whole subtree
    assert deeper_spec.replace({"foo": {"baz": None}}) == Spec(
        {"foo": {"quaz": ["c"]}, "bar": ["b"]}
    )
    assert deeper_spec.replace({"foo": {"baz": None, "quaz": None}}) == Spec(
        {"bar": ["b"]}
    )

    # A leaf is treated as a leaf, even when the original tree is a list
    assert sequence_spec.replace([{"foo": ["a", "b"]}]) == Spec(
        [{"foo": ["a", "b"]}, ["b"]]
    )
    assert sequence_spec.replace(["a", "b"]) == Spec(["a", "b"])


def test_update_leaves():
    assert spec.update_leaves(lambda dims: dims + ["c"]) == Spec(
        {"foo": ["a", "b", "c"], "bar": ["b", "c"]}
    )


def test_size():
    assert spec.size({"foo": np.ones([2, 3]), "bar": np.ones([3])}) == {"a": 2, "b": 3}



"Slice data using a :class:`~spekk.spec.Spec`."

from typing import Protocol, Sequence, Union

from spekk.spec import Spec
from spekk.trees import Tree, leaves, update

IndicesT = Union[int, slice, Sequence[int], None]


class Sliceable(Protocol):
    def __getitem__(self, indices: IndicesT):
        ...


_all_slice = slice(None, None, None)


def slice_array_1(arr: Sliceable, axis: int, indices: IndicesT):
    """Select the indices along the given axis of an array.

    >>> import numpy as np
    >>> arr = np.array([[1,2,3], [4,5,6]])
    >>> slice_array_1(arr, 0, 0)
    array([1, 2, 3])
    >>> slice_array_1(arr, 1, [0, 2])
    array([[1, 3],
           [4, 6]])
    >>> slice_array_1(arr, None, [0, 2])
    array([[1, 2, 3],
           [4, 5, 6]])
    """
    if axis is not None:
        slices = tuple([_all_slice for _ in range(axis)] + [indices])
        arr = arr.__getitem__(slices)
    return arr


def slice_data(
    data: Tree, spec: Spec, slice_definitions: Sequence[Union[str, IndicesT]]
):
    """Slice the data given a spec and a dict (kwargs) of slice definitions. The keys
    of ``slice_definitions`` are the names of the dimensions to slice, and the values 
    are the indices to slice along that dimension. Returns the sliced data and the
    (possibly modified) spec.

    The data may have an arbitrary tree-like shape (nested ``dict``, ``list``, 
    ``tuple``, etc.), but the leaves must support numpy indexing. The returned sliced 
    data has the same shape as the input data. The spec argument describes the shape of 
    the data.

    If the indices for a dimension in ``slice_definitions`` is just an integer, said
    dimension is removed from the returned spec.

    >>> import numpy as np
    >>> data = {"foo": ({"bar": np.ones((4,5))},)}
    >>> spec = Spec({"foo": ({"bar": ["a", "b"]},)})

    Getting the first item in the ``"a"`` dimension removes that dimension from the 
    spec:

    >>> slice_data(data, spec, ("a", 0))
    {'foo': ({'bar': array([1., 1., 1., 1., 1.])},)}

    We can slice multiple dimensions at once:

    >>> slice_data(data, spec, ("a", 0, "b", slice(0, 3)))
    {'foo': ({'bar': array([1., 1., 1.])},)}

    Here, both the ``"a"`` and ``"b"`` dimensions are removed from the spec:

    >>> slice_data(data, spec, ("a", 0, "b", 0))
    {'foo': ({'bar': 1.0},)}
    """
    is_axis = lambda x: isinstance(x, int) or x is None
    for dimension, indices in zip(slice_definitions[::2], slice_definitions[1::2]):
        for leaf in leaves(spec.index_for(dimension), is_axis):
            data = update(
                data, lambda a: slice_array_1(a, leaf.value, indices), leaf.path
            )
        if isinstance(indices, int):
            spec = spec.remove_dimension(dimension)
    return data


def slice_spec(spec: Spec, slice_definitions: Sequence[Union[str, IndicesT]]):
    for dimension, indices in zip(slice_definitions[::2], slice_definitions[1::2]):
        if isinstance(indices, int):
            spec = spec.remove_dimension(dimension)
    return spec


if __name__ == "__main__":
    import doctest

    doctest.testmod()



""":class:`ForAll` transforms a function that works on scalar inputs such that it works 
on arrays instead (vectorization), and can be used with :func:`jax.vmap`."""

from typing import Callable, Optional, Sequence

from spekk import Spec, trees, util
from spekk.transformations import Transformation, common

T_in_axes = Sequence[Optional[int]]
T_vmap = Callable[[callable, T_in_axes], callable]


class ForAll(Transformation):
    """Vectorize/"make looped" a function such that it works on arrays instead of
    scalars.

    >>> from spekk.transformations import ForAll, compose
    >>> f = lambda x, y: x + y
    >>> data =       {"x": range(2), "y": range(3)}
    >>> spec = Spec( {"x": ["dim1"], "y": ["dim2"]} )

    Transform f to run on a grid defined by the "dim1" and "dim2" dimensions:

    >>> tf = compose(f, ForAll("dim2"), ForAll("dim1")).build(spec)
    >>> tf.output_spec
    Spec(['dim1', 'dim2'])
    >>> result = tf(**data)
    >>> util.shape(result)
    (2, 3)
    >>> result
    [[0, 1, 2], [1, 2, 3]]

    You can also use vmap over multiple dimensions at once. Note that the order of the 
    dimensions in ``ForAll("dim2"), ForAll("dim1")`` and ``ForAll("dim1", "dim2")`` is 
    reversed:

    >>> tf = compose(f, ForAll("dim1", "dim2")).build(spec)
    >>> tf(**data)  # This results in the same as in the previous example
    [[0, 1, 2], [1, 2, 3]]"""

    def __init__(
        self,
        dimension: str,
        *additional_dimensions: str,
        vmap_impl: Optional[T_vmap] = None,
    ):
        self.dimensions = [dimension, *additional_dimensions]
        self.vmap_impl = vmap_impl

    def transform_function(
        self, to_be_transformed: callable, input_spec: Spec, output_spec: Spec
    ) -> callable:
        transformed = to_be_transformed
        remaining_dimensions = set(self.dimensions)
        for dimension in reversed(self.dimensions):
            if not input_spec.has_dimension(dimension):
                raise ValueError(f"Spec does not contain the dimension {dimension}.")
            remaining_dimensions.remove(dimension)
            transformed = specced_vmap(
                transformed,
                input_spec.remove_dimension(remaining_dimensions),
                dimension,
                self.vmap_impl,
            )
        return transformed

    def transform_input_spec(self, spec: Spec) -> Spec:
        # The returned function works on 1 element of the dimension at a time, so it has
        # one less dimension.
        return spec.remove_dimension(self.dimensions)

    def transform_output_spec(self, spec: Spec) -> Spec:
        # The dimensions are re-added after vmapping the wrapped function.
        return spec.update_leaves(lambda dimensions: [*self.dimensions, *dimensions])

    def __repr__(self) -> str:
        dimensions_str = ", ".join([repr(dim) for dim in self.dimensions])
        return f"ForAll({dimensions_str})"


def specced_vmap(
    f: callable,
    spec: Spec,
    dimension: str,
    vmap_impl: Optional[T_vmap] = None,
):
    """Similar to ``vmap``, but flattens/decomposes the ``kwargs`` to a list that is
    supported by ``vmap``.
    """
    if vmap_impl is None:
        vmap_impl = python_vmap

    def wrapped(*_unsupported_positional_args, **kwargs):
        if _unsupported_positional_args:
            raise ValueError(
                "Positional arguments are not supported in specced_vmap. Use keyword arguments instead."
            )

        flattened_args, in_axes, unflatten = util.flatten(kwargs, spec, dimension)

        def f_with_unflattening_args(*args):
            original_kwargs = unflatten(args)
            return f(**original_kwargs)

        vmapped_f = vmap_impl(f_with_unflattening_args, in_axes)
        return vmapped_f(*flattened_args)

    return wrapped


def python_vmap(f, in_axes):
    """A simple Python implementation of JAX's :func:`jax.vmap` based on for-loops."""

    def wrapped(*args):
        sizes = [util.shape(arg)[a] for arg, a in zip(args, in_axes) if a is not None]
        size = sizes[0]
        if not all(s == size for s in sizes):
            raise ValueError(
                f"Cannot apply python_vmap to arguments with different sizes over the \
in_axes: {sizes=}, {in_axes=}"
            )

        # The result for each item in the dimension.
        all_results = [
            f(*common.get_args_for_index(args, in_axes, i)) for i in range(size)
        ]
        result0 = all_results[0]

        # Combine the results such that the returned object has the same shape as each
        # individual result.
        combined_result = result0
        for leaf in trees.leaves(
            result0, lambda x: isinstance(x, list) or not trees.has_treedef(x)
        ):
            values = [trees.get(_result, leaf.path) for _result in all_results]
            combined_result = trees.set(combined_result, values, leaf.path)
        return combined_result

    return wrapped


if __name__ == "__main__":
    import doctest

    doctest.testmod()


":class:`TreeLens` is a functional interface to a tree-like data structure."

from typing import Any, Callable, Optional, Sequence, Tuple, TypeVar, Union

from spekk.trees.core import filter, remove, set, traverse, update, update_leaves
from spekk.trees.registry import Tree, treedef

TSelf = TypeVar("TSelf", bound="TreeLens")


class TreeLens:
    """A functional interface to a Tree.

    A lens is an object in functional programming (FP) that allows you to access and
    modify a value in a nested structure in an immutable way.
    """

    def __init__(self, tree: Tree = ()):
        # Ensure that there are no nested TreeLens objects:
        for t in traverse(tree, self.is_leaf):
            if isinstance(t.value, TreeLens):
                tree = set(tree, t.value.tree, t.path)
        self.tree = tree

    def __getitem__(self: TSelf, path: Union[Any, Tuple[Any]]) -> TSelf:
        """Get the value or subtree at the given path.

        If you want to index by a single tuple, e.g. you have a dict with tuples as
        keys, then you should use get(…) instead. This is because there is no way of
        distinguishing between a single tuple argument and a tuple of multiple arguments
        passed to __getitem__(…).
        >>> key = ("a", "tuple", "key")
        >>> tree = TreeLens({key: 1, "a": {"tuple": {"key": "nested_value"}}})
        >>> tree[key]   # This will incorrectly return the nested value
        TreeLens(nested_value)
        >>> tree[key,]  # Adding a comma helps Python distinguish between the two cases
        TreeLens(1)
        >>> tree.get([key])  # get(…) works as expected
        TreeLens(1)
        """
        if not isinstance(path, tuple):
            # NOTE: If you actually want to index by a tuple, e.g. you have a dict with
            # tuples as keys, use get() instead
            path = (path,)
        return self.get(path)

    def get(self: TSelf, path: Sequence[Any]) -> TSelf:
        """Get the value or subtree at the given path.

        >>> tree = TreeLens({"a": {"b": [1, 2, 3]}, "d": [3]})
        >>> tree.get(["a", "b"])
        TreeLens([1, 2, 3])
        >>> tree.get(["a", "b", 1])
        TreeLens(2)
        """
        tree = self.tree
        for k in path:
            if k not in treedef(tree).keys():
                return self.copy_with(None)  # Return with None tree
            tree = tree[k]
        return self.copy_with(tree)

    def has_subtree(self, path: Sequence[Any]) -> bool:
        """Return True if the given path exists in the tree.

        >>> tree = TreeLens({"a": {"b": [1, 2, 3]}})
        >>> tree.has_subtree(["a", "b", 1])
        True
        >>> tree.has_subtree(["a", "c"])
        False
        """
        tree = self.tree
        for k in path:
            if k in treedef(tree).keys():
                tree = tree[k]
            else:
                return False
        return True

    def set(self: TSelf, value: Any, path: Sequence[Any]) -> TSelf:
        """Set the value or subtree at the given path.

        >>> tree = TreeLens({"a": {"b": [1, 2, 3]}, "d": [3]})
        >>> tree.set(5, ["a", "b", 1])
        TreeLens({'a': {'b': [1, 5, 3]}, 'd': [3]})
        """
        return self.copy_with(set(self.tree, value, path))

    def update_subtree(self: TSelf, f: Callable, path: Sequence[Any]) -> TSelf:
        """Update the value or subtree at the given path.

        >>> tree = TreeLens({"a":{"b": [1, 2, 3]}, "d":[3]})
        >>> tree.update_subtree(lambda x: x + 10, ["a", "b", 1])
        TreeLens({'a': {'b': [1, 12, 3]}, 'd': [3]})
        """
        return self.copy_with(update(self.tree, f, path))

    def update_leaves(self: TSelf, f: Callable, path: Sequence[Any] = ()) -> TSelf:
        "See :func:`~spekk.trees.core.update_leaves`."
        return self.copy_with(update_leaves(self.tree, self.is_leaf, f, path))

    def remove_subtree(self: TSelf, path: Sequence[Any]) -> TSelf:
        """Remove the value or subtree at the given path.

        >>> tree = TreeLens({"a": {"b": [1, 2, 3]}, "d": [3]})
        >>> tree.remove_subtree(["a", "b", 1])
        TreeLens({'a': {'b': [1, 3]}, 'd': [3]})
        """
        return self.copy_with(remove(self.tree, path))

    def is_leaf(self, tree: Tree) -> bool:
        """Return True if the given tree is a leaf.

        By default, this is True if the tree has not been registered with the treedef
        registry, but should be overridden for more specialized trees."""
        try:
            treedef(tree)
            return False
        except ValueError:
            return True

    def prune_empty_branches(
        self: Union[TSelf, Tree],
        is_leaf: Optional[Callable[[Tree], bool]] = None,
    ) -> Union[TSelf, Tree]:
        """Remove all empty subtrees.

        May be called as a static method where self is a Tree."""
        if is_leaf is None:
            is_leaf = self.is_leaf
        tree = self.tree if isinstance(self, TreeLens) else self
        not_empty = lambda tree: is_leaf(tree) or len(tree) > 0
        pruned_tree = filter(tree, is_leaf, not_empty)
        if isinstance(self, TreeLens):
            pruned_tree = self.copy_with(pruned_tree)
        return pruned_tree

    def __iter__(self):
        return iter([self.copy_with(t) for t in self.tree])

    def copy_with(self: TSelf, tree: Tree) -> TSelf:
        "Return a copy of this object with the given tree."
        return self.__class__(tree)

    def __repr__(self):
        return f"TreeLens({self.tree})"


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import functools

import jax
import jax.numpy as jnp

from spekk import Spec
from spekk.transformations import *


def kernel_fn(a: float, b: float, c: float):
    return a * b + a * c[0] + b * c[1]


data = {
    "a": jnp.ones((2, 3, 4)),
    "b": jnp.ones((4, 2)),
    "c": [jnp.ones((2, 5)), jnp.ones((3,))],
}
spec = Spec(
    {
        "a": ["a", "b", "c"],
        "b": ["c", "a"],
        "c": (["a", "d"], ["b"]),
    }
)

p = compose(
    kernel_fn,
    ForAll("c", jax.vmap),
    ForAll("a", jax.vmap),
    ForAll("b", jax.vmap),
    Reduce.Sum("d", reduce_impl=functools.reduce),
    Apply(jnp.sum, Axis("b")),
    Wrap(jax.jit),
).build(spec)
result = p(**data)
print(result.shape)



"Some common utility functions used by :mod:`spekk.transformations`."

from typing import Any, Sequence, Union

import numpy as np

from spekk import trees
from spekk.transformations import common


def compose(x, *wrapping_functions):
    """Apply each f in fs to x.

    Let's say we have some functions:

    >>> f = lambda x: x+1
    >>> g = lambda x: x*2
    >>> h = lambda x: x**2

    We can use :func:`compose` to apply each function in order:

    >>> compose(1, f, g, h)  # ((1 + 1) * 2) ** 2 = 16
    16

    This would be the same as calling:

    >>> h(g(f(1)))  # ((1 + 1) * 2) ** 2 = 16
    16

    In situations with a lot of nested function calls, :func:`compose` may be more
    readable. Also notice that when using compose, functions are evaluated in the order
    that they are passed in (left-to-right), while with the nested function calls, the
    functions are evaluated in the reverse order (right-to-left).

    :func:`compose` can also be used to build up a function from smaller function
    transformations:

    >>> wrap_f_double = lambda f: (lambda x: 2*f(x))
    >>> wrap_f_square = lambda f: (lambda x: f(x)**2)
    >>> f = compose(
    ...   lambda x: x+1,
    ...   wrap_f_double,
    ...   wrap_f_square,
    ... )
    >>> f(1)  # ((1 + 1) * 2) ** 2 = 16
    16
    """
    for wrap in wrapping_functions:
        x = wrap(x)
    return x


def identity(x):
    "Return the input unchanged."
    return x


def get_fn_name(f) -> str:
    if hasattr(f, "__qualname__"):
        return f.__qualname__
    if hasattr(f, "__name__"):
        return f.__name__
    return repr(f)


def getitem_along_axis(x, axis: int, i: int):
    slice_ = tuple([slice(None)] * axis + [i])
    try:
        return x.__getitem__(slice_)
    except TypeError:
        try:
            return np.array(x).__getitem__(slice_)
        except Exception:
            raise ValueError(
                f"Cannot get item at index {i} along axis {axis} for {x!r}"
            )


def get_args_for_index(
    args: Sequence, in_axes: Sequence[Union[int, None]], i: int
) -> Sequence:
    return [
        common.getitem_along_axis(arg, a, i) if a is not None else arg
        for arg, a in zip(args, in_axes)
    ]


def map_1_flattened(
    map_f: callable,
    flattened_args: Sequence[Any],
    in_axes: Sequence[Union[int, None]],
    unflatten: callable,
    i: int,
):
    args = []
    for arg, axis in zip(flattened_args, in_axes):
        # If axis is None then we leave the argument as is.
        if axis is not None:
            # If axis is not None, then get the item at index i along the given axis.
            arg = trees.update_leaves(
                arg,
                lambda x: not trees.has_treedef(x),
                lambda x: common.getitem_along_axis(x, axis, i),
            )
        args.append(arg)

    kwargs = unflatten(args)
    return map_f(**kwargs)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python
from setuptools import find_packages, setup

setup(
    name="spekk",
    version="1.0.3",
    description="Named dimensions of nested arrays, independent of array implementation (Numpy, JAX, etc)",
    author="Magnus Dalen Kvalevåg",
    author_email="magnus.kvalevag@ntnu.no",
    packages=find_packages(),
    install_requires=["numpy"],
)



""":class:`Wrap` is a simple :class:`Transformation` that wraps a function with another 
function, such as :func:`jax.jit`."""

from spekk import Spec
from spekk.transformations import Transformation, common


class Wrap(Transformation):
    """Simply wraps a function with another function, but useful for keeping 
    information about the spec in a chain of :class:`Transformation`.
    
    Attributes:
        f: A wrapper function, for example :func:`jax.jit`.
        args: Optional extra positional arguments to pass to ``f``.
        kwargs: Optional extra keyword arguments to pass to ``f``.

    Example:
        >>> import jax
        >>> my_fn = lambda x: x**2
        >>> wrapped_fn1 = Wrap(jax.jit)(my_fn)
        >>> wrapped_fn2 = jax.jit(my_fn)

        `wrapped_fn1` and `wrapped_fn2` are equivalent, but `wrapped_fn1` will propagate 
        information about the spec (if applicable) to nested :class:`Transformation`.
    """
    def __init__(self, f: callable, *args, **kwargs):
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def transform_function(
        self, to_be_wrapped: callable, input_spec: Spec, output_spec: Spec
    ) -> callable:
        return self.f(to_be_wrapped, *self.args, **self.kwargs)

    def transform_input_spec(self, spec: Spec) -> Spec:
        return spec

    def transform_output_spec(self, spec: Spec) -> Spec:
        return spec

    def __repr__(self) -> str:
        args_str = ", ".join([str(arg) for arg in self.args])
        kwargs_str = ", ".join([f"{k}={str(v)}" for k, v in self.kwargs.items()])
        repr_str = f"Wrap({common.get_fn_name(self.f)}"
        if self.args:
            repr_str += f", {args_str}"
        if self.kwargs:
            repr_str += f", {kwargs_str}"
        # Make sure the repr string is not too long
        if len(repr_str) > 140:
            repr_str = repr_str[: (140 - len("… <truncated>"))] + "… <truncated>"
        return repr_str + ")"



"""Transformations are higher-order functions used to wrap other functions in order to 
modify their behavior. :mod:`spekk.transformations` additionally allows you to specify
how the spec changes when a function is transformed.

Example:

    Let's say we have a function ``f`` that takes a single scalar argument ``x`` and 
    returns ``x**2``:

    >>> f = lambda x: x**2

    Let's say we also have some data and spec where ``x`` is a list of numbers:

    >>> from spekk import Spec
    >>> data = {"x": [1, 2, 3]}
    >>> spec = Spec({"x": ["numbers"]})

    We can use :class:`spekk.transformations.ForAll` to transform ``f`` into a function
    that takes a list of numbers and returns a list of numbers:

    >>> from spekk.transformations import ForAll
    >>> tf = ForAll("numbers")(f)
    >>> tf = tf.build(spec)  # Let the transformation know what the spec of the data is
    >>> tf(**data)
    [1, 4, 9]

    Now we have transformed ``f`` to take a list of integers instead of just a number!
    If we are using JAX, we could use :class:`spekk.transformations.ForAll` along with
    :func:`jax.vmap` to get the same result, except vectorized and running in parallel 
    on a GPU.

    This may still seem very verbose, but imagine that ``f`` is a more complex function 
    that takes in many more arguments, where some of the arguments have some of the 
    same dimensions. We can build up functionality by composing transformations:

    >>> from spekk.transformations import compose, Apply, Axis
    >>> import jax.numpy as jnp
    >>> import jax
    >>> some_complex_function = lambda *args, **kwargs: ...  # Just imagine it...
    >>> some_complex_spec = Spec()  # Just imagine it...
    >>> tf = compose(
    ...     some_complex_function,
    ...     ForAll("dimension0", vmap_impl=jax.vmap),
    ...     ForAll("dimension1", vmap_impl=jax.vmap),
    ...     Apply(jnp.sum, Axis("dimension1")),
    ...     ForAll("dimension2", vmap_impl=jax.vmap),
    ... ).build(some_complex_spec)
    
On a high level, a transformation takes in a function as an argument, the spec of the 
input-arguments, and the spec of the returned value from the function, and returns a 
new transformed function (of type :class:`spekk.transformations.TransformedFunction`).
It also gives information about how the spec of the original function changes when 
transformed.

A schematic of how :mod:`spekk.transformations` transforms a function is shown below. 
Note how it can be arbitrarily stacked to build up functionality.

.. code-block:: text

    input_spec         Transformed function         output_spec
        │                        ▲                        ▲
    ┌───┼────────────────────────┼────────────────────────┼───┐
    │   ├──────────────┐         │            ┌───────────┴─┐ │  
    │   │              │  ┌──────┴──────┐     │ Transform   │ │
    │ ┌─▼───────────┐  │  │ Transform   │     │ output spec │ │
    │ │ Transform   │  │  │ function    │     └──────▲──────┘ │
    │ │ input spec  │  │  └─▲────▲────▲─┘            │        │
    │ └──────┬──────┘  └────┘    │    └──────────────┤        │
    └────────┼───────────────────┼───────────────────┼────────┘
             │                   │                   │
        passed_spec              │              returned_spec
             │                   │                   │
           ┌╶▼╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶┴╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶╶┴╶┐
           ╎            Wrapped function               ╎
           └╶╶╶╶╶╶┬╶╶╶╶╶╶╶╶╶╶╶╶╶╶▲╶╶╶╶╶╶╶╶╶╶╶╶╶╶▲╶╶╶╶╶╶┘
                  ╎              ╎              ╎

For example, :class:`spekk.transformations.ForAll` transforms a function such that it 
runs for each element of the input data (along a dimension named by spec) individually. 
It can be used to transform a function such that it works on lists of numbers instead 
of just a single number, similarly to :func:`jax.vmap`.

Let's call the function that is transformed the "wrapped function.". Because the 
wrapped function runs on each element individually, we remove the looped-over/vecorized 
dimension from the spec before it is passed on down. Thus, 
:class:`spekk.transformations.ForAll` removes the dimension when transforming the input 
spec (visualied as the left-most inner rectangle in the above schematic).

The transformed input spec is named ``passed_spec`` in the schematic, and is passed 
down to the wrapped function. The wrapped function may be a 
:class:`TransformedFunction` itself, in which case it also has information about the 
spec of the returned value. The spec of the returned value is called ``returned_spec`` 
in the schematic.

In the case of :class:`spekk.transformations.ForAll`, the looped-over/vecorized 
dimension is re-added to the spec before being returned. The final returned spec is 
called ``output_spec`` in the schematic.

The actual transformation step takes the wrapped function, the original ``input_spec``, 
and the ``returned_spec`` to produce a new function.
"""

from spekk.transformations import common
from spekk.transformations.apply import Apply
from spekk.transformations.axis import Axis
from spekk.transformations.base import (
    PartialTransformation,
    Specced,
    Transformation,
    TransformedFunction,
)
from spekk.transformations.common import compose
from spekk.transformations.for_all import ForAll
from spekk.transformations.reduce import Reduce
from spekk.transformations.wrap import Wrap

__all__ = [
    "common",
    "Apply",
    "Axis",
    "PartialTransformation",
    "Specced",
    "Transformation",
    "TransformedFunction",
    "compose",
    "ForAll",
    "Reduce",
    "Wrap",
]


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import hypothesis as ht
import numpy as np
from test_helpers.generators.spec import kwargs_specs
from test_helpers.mock_data import generate_mock_data

from spekk import Spec, trees
from spekk.transformations import reduce


def _add_leaves(o1: trees.Tree, o2: trees.Tree):
    state = o1
    for leaf in trees.leaves(state, lambda x: isinstance(x, (np.ndarray, float, int))):
        state = trees.set(state, leaf.value + trees.get(o2, leaf.path), leaf.path)
    return state


def _sum_tree_with_numpy(tree: trees.Tree, spec: Spec, dimension: str):
    dim_size = spec.size(tree, dimension)
    for leaf in trees.leaves(spec.index_for(dimension), is_axis):
        axis = leaf.value
        if axis is None:
            # If the axis is None it will be added to itself for each iteration, aka
            # dim_size times.
            tree = trees.update(tree, lambda x: x * dim_size, leaf.path)
        else:
            # If the axis is an integer, it represents the dimension and is summed over.
            tree = trees.update(tree, lambda x: np.sum(x, axis), leaf.path)
    return tree


is_axis = lambda x: isinstance(x, int) or x is None
is_array_leaf = lambda x: isinstance(x, (np.ndarray, float, int))


@ht.given(kwargs_specs())
def test_reduce_addition_is_same_as_sum(spec: Spec):
    data = generate_mock_data(spec)
    for dimension in spec.dimensions:
        reduce_result = reduce.specced_map_reduce(
            dict, _add_leaves, data, spec, dimension
        )
        sum_result = _sum_tree_with_numpy(data, spec, dimension)
        assert trees.are_equal(reduce_result, sum_result, is_array_leaf, np.array_equal)
